# import lmfit parameters
from .minimizer import Parameter, Parameters

# Import models
from .ordinary import EpistasisNonlinearRegression
from .power import EpistasisPowerTransform
from .spline import EpistasisSpline
